# -- import packages: ---------------------------------------------------------
import ABCParse
import pandas as pd
import pathlib
import scdiffeq as sdq


# -- set typing: --------------------------------------------------------------
from typing import List


# -- operational object class: ------------------------------------------------
class VersionAccuracy(ABCParse.ABCParse):
    def __init__(
        self,
        version: sdq.io.Version,
        train_col: str = "unique_train.all_fates",
        test_col: str = "unique_test.all_fates",
        *args,
        **kwargs,
    ):
        self.__parse__(locals())

    @property
    def csv_paths(self) -> List[pathlib.Path]:
        return list(self._version._PATH.glob("fate_prediction_metrics/*/accuracy.csv"))

    def _read_frame(self, path: pathlib.Path):

        name = path.parent.name
        return (
            pd.read_csv(path, index_col=0)
            .loc[[self._train_col, self._test_col]]
            .rename({self._train_col: "train", self._test_col: "test"}, axis=0)
            .rename({"accuracy": name}, axis=1)
        )

    def _format_ckpt_name(self, name):
        return name.replace("=", "_").split(".ckpt")[0].replace("-", ".")

    @property
    def _saved_ckpt_fpaths(self) -> pathlib.Path:
        return [ckpt.path for ckpt in self._version.ckpts.values()]

    @property
    def _saved_ckpts(self):
        """Not all (e.g., `on_train_epoch_end`) ckpts get saved. Use this to filter accordingly"""
        return [self._format_ckpt_name(ckpt.name) for ckpt in self._saved_ckpt_fpaths]

    def _get_ckpt_epoch_values(self, df):
        ckpt_epochs = df.index.str.split(".", expand=True).to_frame()[0].values.tolist()
        epochs = []
        for epoch in ckpt_epochs:
            if epoch == "last":
                epochs.append(2500)
            else:
                epochs.append(int(epoch.split("epoch_")[1]))

        return epochs

    @property
    def df(self):
        if not hasattr(self, "_df"):
            df = pd.concat(
                [self._read_frame(path) for path in self.csv_paths], axis=1
            ).T
            df = df.loc[df.index.isin(self._saved_ckpts)]
            df["ckpt_path"] = self._saved_ckpt_fpaths
            df["epoch"] = self._get_ckpt_epoch_values(df)
            self._df = df.sort_values(["train", "epoch"], ascending=[False, False])
        return self._df

    @property
    def best_training_ckpt(self):
        return self.df.index[0]

    @property
    def best_test_from_train(self) -> pd.Series:
        return self.df.loc[self.best_training_ckpt]



# -- import packages: ----------------------------------------------------------
import anndata
import numpy as np
import pathlib
import scdiffeq as sdq
import torch

# -- set type hints: ----------------------------------------------------------
from typing import Union


# -- define function: ----------------------------------------------------------
def annotate_adata_with_growth_weights(
    adata: anndata.AnnData,
    growth_weights_path: Union[pathlib.Path, str],
    weight_key: str = "W",
    time_key: str = "Time point",
    time_point: int = 2,
):

    weight_pt = sdq.datsets.larry_growth_weights()
    # weight_pt = torch.load(growth_weights_path, weights_only=False)
    W = np.ones(len(adata))
    W[adata.obs[time_key] == time_point] = weight_pt["w"][0]
    adata.obs[weight_key] = W
    return adata



import numpy as np
import matplotlib.cm as cm
class BoxPlot:
    def __init__(self, boxplot_kwargs: dict, colors: list = cm.tab20.colors) -> None:

        self.boxplot_kwargs = boxplot_kwargs
        self.colors = colors
    @property
    def y(self):
        return list(self.data.values())

    @property
    def x(self):
        return np.arange(len(self.y)) + 1

    def forward(self):
        bp = self.ax.boxplot(
            self.y,
            positions=self.x,
            patch_artist=True,
            **self.boxplot_kwargs,
        )
        if self.mode == "background":
            bp = self._format_background(bp)
        elif self.mode == "foreground":
            bp = self._format_foreground(bp)

        bp = self._format_single_data_point(bp)

        return bp

    def _format_background(self, bp):

        for median in bp["medians"]:
            median.set_visible(False)
        for en, mean in enumerate(bp["means"]):
            mean.set_c(self.colors[en])

        for en, box in enumerate(bp["boxes"]):
            box.set_facecolor(self.colors[en])
            box.set_alpha(0.2)

        for en, whisker in enumerate(bp["whiskers"]):
            whisker.set_c("None")

        for en, cap in enumerate(bp["caps"]):
            cap.set_c("None")

        return bp

    def _format_foreground(self, bp):
        for en, box in enumerate(bp["boxes"]):
            box.set_facecolor("None")
            box.set_edgecolor(self.colors[en])

        colors_ = np.repeat(
            np.array(self.colors), 2, axis=0
        )  # list(np.repeat(self.colors, 2))
        for en, whisker in enumerate(bp["whiskers"]):
            whisker.set_c(colors_[en])

        for en, cap in enumerate(bp["caps"]):
            cap.set_c(colors_[en])

        for median in bp["medians"]:
            median.set_visible(False)

        return bp

    def _format_single_data_point(self, bp):
        # Remove boxes for categories with single data points
        for en, (key, val) in enumerate(self.data.items()):
            if len(val) == 1:
                for bp_key in bp.keys():
                    if len(bp[bp_key]) > 0:
                        if len(bp[bp_key]) == len(self.data):
                            bp[bp_key][en].set_visible(False)
                        elif len(bp[bp_key]) == len(self.data) * 2:
                            bp[bp_key][2 * en].set_visible(False)
                            bp[bp_key][2 * en + 1].set_visible(False)
        return bp

    def __call__(self, ax, data, mode: str = "background"):

        self.ax = ax
        self.data = data
        self.mode = mode

        return self.forward()



# -- import packages: ---------------------------------------------------------
import pathlib

# -- import local dependencies: -----------------------------------------------
from ._model_runner import Runner

# -- import external dependencies: --------------------------------------------
from typing import Dict, List, Optional, Union


# -- define function: --------------------------------------------------------
def run_fate_prediction(
    project_name: str,
    h5ad_path: str,
    time_key: str,
    growth_weights_path: Optional[Union[pathlib.Path, str]] = None,
    weight_key: str = "W",
    ckpt_frequency: int = 1,
    save_last_ckpt: bool = True,
    keep_ckpts: int = 3,
    monitor: str = "epoch_validation_loss",
    n_seeds: int = 1,
    seeds: List[int] = [0],
    n_eval: int = 2000,
    swa_lrs: float = 1e-5,
    mu_hidden: List[int] = [512, 512],
    sigma_hidden: List[int] = [32, 32],
    train_epochs: int = 2500,
    train_lr: float = 4e-5,
    train_step_size: int = 1500,
    mu_dropout: float = 0,
    sigma_dropout: float = 0,
    batch_size: int = 2048,
    coef_diffusion: float = 1,
    diffeq_type: str = "SDE",
    use_key: str = "X_pca",
    potential_type: str = "fixed",
    latent_dim: int = 50,
    velocity_ratio_params: Dict[str, float] = {
        "target": 2.5,
        "enforce": 100,
        "method": "square",
    },
) -> None:
    runner = Runner(
        h5ad_path=h5ad_path,
        project_name=project_name,
        time_key=time_key,
        growth_weights_path=growth_weights_path,
        weight_key=weight_key,
        ckpt_frequency=ckpt_frequency,
        save_last_ckpt=save_last_ckpt,
        keep_ckpts=keep_ckpts,
        monitor=monitor,
    )
    return runner(
        n_seeds=n_seeds,
        seeds=seeds,
        n_eval=n_eval,
        swa_lrs=swa_lrs,
        mu_hidden=mu_hidden,
        sigma_hidden=sigma_hidden,
        train_epochs=train_epochs,
        train_lr=train_lr,
        train_step_size=train_step_size,
        mu_dropout=mu_dropout,
        sigma_dropout=sigma_dropout,
        batch_size=batch_size,
        potential_type=potential_type,
        coef_diffusion=coef_diffusion,
        diffeq_type=diffeq_type,
        use_key=use_key,
        latent_dim=latent_dim,
        velocity_ratio_params=velocity_ratio_params,
    )



# -- import packages: ---------------------------------------------------------
import pandas as pd

# -- operational object cls: --------------------------------------------------
class ScreenResult:
    """Fate perturbation screen result (singular version)"""
    def __init__(self, base_path: str) -> None:
        self._base_path = base_path
        self._suffixes = {
            "lfc_mean": ".lfc.csv",
            "pval": ".pval.csv",
            "lfc_std": ".lfc_std.csv",
        }

    def _fetch_data(self, accessor: str) -> pd.DataFrame:
        if not hasattr(self, f"_{accessor}"):
            csv_path = str(self._base_path) + self._suffixes[accessor]
            df = pd.read_csv(csv_path, index_col=0)
            setattr(self, f"_{accessor}", df)
        return getattr(self, f"_{accessor}")

    @property
    def lfc_mean(self) -> pd.DataFrame:
        return self._fetch_data("lfc_mean")

    @property
    def pval(self) -> pd.DataFrame:
        return self._fetch_data("pval")

    @property
    def lfc_std(self) -> pd.DataFrame:
        return self._fetch_data("lfc_std")



# -- import packages: ----------------------------------------------------------
import anndata
import logging
import pathlib

# -- import local dependencies: -----------------------------------------------
from ._annotate_adata_with_growth_weights import annotate_adata_with_growth_weights

# -- set type hints: ----------------------------------------------------------
from typing import Optional, Union

# -- configure logger: --------------------------------------------------------
logger = logging.getLogger(__name__)

# -- define function: ---------------------------------------------------------
def load_adata(
    h5ad_path: Union[pathlib.Path, str],
    growth_weights_path: Optional[Union[pathlib.Path, str]] = None,
    weight_key: str = "W",
    time_key: str = "Time point",
    time_point: int = 2,
) -> None:
    print("Loading adata...")
    adata = anndata.read_h5ad(h5ad_path, silent=True)

    # -- UPDATE TO TRAIN ONLY ON WELLS 0, 1: ----------------------------
    adata.obs["train"] = adata.obs["Well"].isin([0, 1])

    if not growth_weights_path is None:
        logger.info("Annotating adata with gro    wth weights...")
        adata = annotate_adata_with_growth_weights(
            adata, growth_weights_path, weight_key, time_key, time_point
        )
        logger.info(f"\nAnnData prepared:\n{adata}")
    return adata



from .__version__ import __version__

from . import fate_prediction
from . import parsers
from . import _plotting as pl
from . import _tools as tl
from . import metrics

__all__ = ["fate_prediction", "parsers", "metrics", "pl", "tl", "__version__"]
